"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createConverter = void 0;
const events_1 = require("./events");
const interop_1 = require("./interop");
const infra_codegen_1 = require("../infra.codegen");
const noop = () => { };
const mediaRules = [
    { attribute: 'src', selector: 'img, embed, video > source, video > track, audio > source' },
    { attribute: 'srcset', selector: 'picture > source' },
];
function prefixMediaSources(component, prefix) {
    const prefixAttributeValue = (el, attr) => el.setAttribute(attr, prefix + el.getAttribute(attr));
    for (const { attribute, selector } of mediaRules) {
        Array.from(component.querySelectorAll(selector))
            .filter((el) => el.hasAttribute(attribute) && !el.getAttribute(attribute).match(/^https?:/))
            .forEach((el) => prefixAttributeValue(el, attribute));
    }
}
function project(component, destination, options) {
    if (options?.resourcePathRoot && !infra_codegen_1.default.noMutation) {
        prefixMediaSources(component, options.resourcePathRoot);
    }
    destination.appendChild(component);
}
function makeUrl(href) {
    const origin = document.location.origin;
    if (!href.startsWith(origin)) {
        return `${origin}${href}`;
    }
    return href;
}
function createConverter(lazy, opts, language, logLevel) {
    const bootLoader = (0, interop_1.createBootLoader)(infra_codegen_1.default.url, infra_codegen_1.default.satellites);
    const boot = (opts) => bootLoader(opts).then(async (res) => {
        const [_, capabilities] = res;
        if (capabilities.includes('logging')) {
            if (typeof logLevel === 'number') {
                await (0, interop_1.setLogLevel)(logLevel);
            }
        }
        if (capabilities.includes('events')) {
            const eventDispatcher = document.body.dispatchEvent;
            // listen to all events for forwarding them
            document.body.dispatchEvent = function (ev) {
                if (ev.type.startsWith('piral-')) {
                    const type = ev.type.replace('piral-', '');
                    const args = ev.detail.arg;
                    try {
                        JSON.stringify(args);
                        (0, interop_1.processEvent)(type, args);
                    }
                    catch {
                        console.warn(`The event "${type}" could not be serialized and will not be handled by Blazor.`);
                    }
                }
                return eventDispatcher.call(this, ev);
            };
        }
        if (language && capabilities.includes('language')) {
            if (typeof language.current === 'string') {
                await (0, interop_1.setLanguage)(language.current);
            }
            if (typeof language.onChange === 'function') {
                language.onChange(interop_1.setLanguage);
            }
        }
        window.dispatchEvent(new CustomEvent('loaded-blazor-core'));
        return res;
    });
    let loader = !lazy && boot(opts);
    let listener = undefined;
    const enqueueChange = (locals, update) => {
        if (typeof update !== 'function') {
            // nothing to do in this case
        }
        else if (locals.state === 'mounted') {
            loader.then(update);
        }
        else {
            locals.next = update;
        }
    };
    const convert = (moduleName, dependency, args, options) => ({
        mount(el, data, ctx, locals) {
            const props = { ...args, ...data };
            const { piral } = data;
            const nav = ctx.navigation;
            el.setAttribute('data-blazor-pilet-root', 'true');
            (0, events_1.addGlobalEventListeners)(el);
            locals.state = 'fresh';
            locals.next = noop;
            locals.dispose = (0, events_1.attachEvents)(el, (ev) => {
                ev.stopPropagation();
                const { target, props } = ev.detail;
                piral.renderHtmlExtension(target, props);
            }, (ev) => {
                ev.stopPropagation();
                const { to, state, replace } = ev.detail;
                replace ? nav.replace(to, state) : nav.push(to, state);
            }, (ev) => {
                ev.stopPropagation();
                const { type, args } = ev.detail;
                piral.emit(type, args);
            });
            function mountClassic(config) {
                return (0, interop_1.activate)(moduleName, props).then((refId) => {
                    const [root] = config;
                    const node = root.querySelector(`#${refId} > div`);
                    locals.unmount = () => {
                        root.querySelector(`#${refId}`)?.appendChild(node);
                        (0, interop_1.deactivate)(moduleName, refId);
                        el.innerHTML = '';
                    };
                    locals.update = (props) => {
                        (0, interop_1.reactivate)(moduleName, refId, props);
                    };
                    project(node, el, options);
                });
            }
            function mountModern(_) {
                return (0, interop_1.createElement)(moduleName, props).then((refId) => {
                    const child = document.createElement('piral-blazor-component');
                    child.setAttribute('rid', refId);
                    el.appendChild(child);
                    locals.unmount = () => {
                        (0, interop_1.destroyElement)(refId);
                        child.remove();
                        el.innerHTML = '';
                    };
                    locals.update = (props) => {
                        (0, interop_1.updateElement)(refId, props);
                    };
                });
            }
            (loader || (convert.loader = loader = boot(opts)))
                .then((config) => {
                if (listener === undefined) {
                    listener = nav.listen(({ location, action }) => {
                        // POP is already handled by .NET
                        if (action !== 'POP') {
                            const url = makeUrl(location.href);
                            (0, interop_1.callNotifyLocationChanged)(url, action === 'REPLACE', location.state);
                        }
                    });
                }
                return config;
            })
                .then((config) => dependency(config).then(() => {
                if (locals.state === 'fresh') {
                    const [_, capabilities, applyChanges] = config;
                    const fn = capabilities.includes('custom-element') ? mountModern : mountClassic;
                    applyChanges(piral);
                    return fn(config).then(() => {
                        locals.state = 'mounted';
                        locals.next(config);
                        locals.next = noop;
                    });
                }
            }))
                .catch((err) => console.error(err));
        },
        update(el, data, ctx, locals) {
            enqueueChange(locals, () => {
                locals.update?.({ ...args, ...data });
            });
        },
        unmount(el, locals) {
            (0, events_1.removeGlobalEventListeners)(el);
            el.removeAttribute('data-blazor-pilet-root');
            locals.dispose();
            enqueueChange(locals, locals.unmount);
            locals.state = 'removed';
        },
    });
    convert.boot = () => boot(opts);
    convert.loader = loader;
    convert.lazy = lazy;
    return convert;
}
exports.createConverter = createConverter;
//# sourceMappingURL=converter.js.map